"""Utility functionality used within the `e2e.*` ecosystem."""
from typing import Any
from typing import List
from typing import Type
from typing import TypeVar

T = TypeVar("T")


def fqualname_of(obj: Any) -> str:
    """Gets the fully-qualified name for the given object."""
    return "{}.{}".format(obj.__class__.__module__, obj.__class__.__qualname__)


def fname_of(obj: Any) -> str:
    """Get the module-qualified path for the class.

    This will not properly communicate nested classes, functions, etc.
    """
    return "{}.{}".format(obj.__class__.__module__, obj.__class__.__name__)


def subclasses_of(cls: Type[T]) -> List[Type[T]]:
    """Fetch the whole subclass tree of the given class as a list.

    No ordering guarantee is provided within the list.
    """
    all_subclasses = []

    for subclass in cls.__subclasses__():
        all_subclasses.append(subclass)
        all_subclasses.extend(subclasses_of(subclass))

    return all_subclasses



"""Base functionality for modelling frameworks."""
__all__ = ["NamedParentable", "ParentableMeta"]


import functools
import types
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import TypeVar

from . import util

T = TypeVar("T")
T_P = TypeVar("T_P", bound="NamedParentable")


def make_prop(original_element: T_P) -> property:
    """Wraps the given object into a `property`."""

    def ret_contained_element(self: "NamedParentable") -> T_P:
        # FIXME: switch to using a ._parent_override here.
        # if original_element.parent is None:
        original_element.parent = self
        return original_element

    return property(ret_contained_element)


def make_prop_via(original_element: T_P, parent_property: property) -> property:
    """Wraps the given object into a `property`, assigning the given parent.

    The given parent must be a `property` object, i.e. `Class.property` and not
    `object.property` (since the latter would run and return from the property).
    """

    def ret_contained_element(self: "NamedParentable") -> T_P:
        original_element.parent = parent_property.__get__(self, type(self))
        return original_element

    return property(ret_contained_element)


def stitch_parent(func: Callable[..., T]) -> Callable[..., T]:
    """Decorate a method on a model to stitch the parent on return."""

    @functools.wraps(func)
    def wrapper(self: "NamedParentable", *args: Any, **kwargs: Any) -> T:
        ret = func(self, *args, **kwargs)
        if ret != self and isinstance(ret, NamedParentable):
            ret.parent = self
        return ret

    return wrapper


class ParentableMeta(type):
    """Metaclass which handles stitching parents into a parented class model.

    Basic (parent not specified) model constructs are wrapped in a `property`
    which sets the parent to the containing class.

    More complex (user-specified parent) constructs are also wrapped in a
    `property`, but their parent is assigned as other basic or complex
    properties. These parents are bound and computed on access.
    """

    # An improvement would be to generate these property mappings, get all their
    # values once, and use that as the model. This could use some caching
    # mechanism or perhaps render out a new instance, which likely requires the
    # implementing class to provide this behaviour on ``__new__``.

    def __new__(
        cls, name: str, bases: Tuple[type, ...], namespace: Dict[str, Any]
    ) -> Any:
        known_model_objects: Dict[str, Any] = {}
        known_model_properties: Dict[Any, property] = {}

        # First pass: Record known model constructs and initial property proxies
        for attr in namespace:
            candidate = namespace[attr]
            # if type(candidate).__class__ is cls:
            # FIXME: Is `cls` appropriate here? Probably not? Multi metaclass?
            if isinstance(candidate.__class__, cls):
                known_model_objects[attr] = candidate
                known_model_properties[candidate] = make_prop(candidate)

        # Second pass: Remap user-given parents, finalize other model constructs
        # FIXME: TBD if this depends on `namespace` being ordered. Probably.
        for attr in namespace:
            candidate = namespace[attr]
            # if type(candidate).__class__ is cls:
            if isinstance(candidate.__class__, cls):
                if candidate.parent is not None:
                    parent_prop = known_model_properties[candidate.parent]
                    namespace[attr] = make_prop_via(candidate, parent_prop)
                    known_model_properties[candidate] = namespace[attr]
                else:
                    # Delayed evaluation in case objects remapped above
                    namespace[attr] = known_model_properties[known_model_objects[attr]]

        # Decoration pass: Wrap user-defined methods to point the parent model.
        # This actually wraps the framework's methods too, so we have to be careful.
        for attr in namespace:
            candidate = namespace[attr]
            if isinstance(candidate, types.FunctionType) and not (
                attr.startswith("__") and attr.endswith("__")
            ):
                namespace[attr] = stitch_parent(namespace[attr])

        return super().__new__(cls, name, bases, namespace)


class NamedParentable(metaclass=ParentableMeta):
    """Provides auto-stitched parents for Object-Modelling frameworks.

    The instance containing an instance of this class will be accessible via
    the `parent` attribute, with the whole hierarchy of parents accessible via
    the `parent_chain` property.

    It is possible to have no parent, e.g. for a top-level object.

    This provides additional functionality for labelling objects. The string
    representation will include the parent hierarchy, and this can be utilised
    in logging for effective human-friendly inspection.

    Args:
        name: A human-friendly name, utlised for string representation and
            logging.

    Keyword Args:
        parent: Optional explicit parent override, if desired.
    """

    name: str = "UNKNOWN"
    parent: Optional["NamedParentable"] = None

    def __init__(
        self, name: str, *, parent: Optional["NamedParentable"] = None,
    ):
        self.name = name
        self.parent = parent

    @property
    def parent_chain(self) -> List["NamedParentable"]:
        """Returns a list of parents in order of increasing distance."""
        return [self.parent] + self.parent.parent_chain if self.parent else []

    def __repr__(self) -> str:
        return "{}(name={})".format(util.fqualname_of(self), self.name)

    def __str__(self) -> str:
        return " in ".join([repr(p.name) for p in [self] + self.parent_chain])



"""Various tests to ensure inline (script) modelling works as expected."""

from e2e.common import modelling


def test_basic_modelling() -> None:
    """Ensure basic inline modelling works.

    Without using class-based modelling, `e2e` modelled constructs should be
    entirely functional. This only requires that the user manually supplies the
    parent.
    """

    root = modelling.NamedParentable("Root")
    leaf = modelling.NamedParentable("Leaf", parent=root)

    # Ensure these still behave as their components do
    assert root.name == "Root"
    assert leaf.name == "Leaf"

    # Ensure proper linkage
    assert root.parent is None
    assert leaf.parent == root



"""Various tests to ensure class-based modelling works as expected."""
# pylint: disable=missing-docstring,protected-access

from e2e.common import modelling


def test_basic_class_linkage() -> None:
    """Ensure that class modelling links the containing class as the parent.

    If no parent is specified, the containing instance should be the parent.
    """

    class Model(modelling.NamedParentable):
        leaf = modelling.NamedParentable("Leaf")

    m = Model("TestModel")

    assert m.leaf.parent == m
    assert m.parent is None


def test_hidden_class_linkage() -> None:
    """Ensure that overriding parents via class modelling works as expected.

    If a parent is specified, it should be used.
    """

    class Model(modelling.NamedParentable):
        _hidden = modelling.NamedParentable("Hidden")
        leaf = modelling.NamedParentable("Leaf", parent=_hidden)

    m = Model("TestModel")

    assert m.leaf.parent != m
    assert m.leaf.parent == m._hidden


def test_deeply_hidden_class_linkage() -> None:
    """Ensure that overriding parents via class modelling works as expected.

    If a parent is specified, it should be used, no matter how deep the override
    structure goes.
    """

    class Model(modelling.NamedParentable):
        _hidden_top = modelling.NamedParentable("_Top")
        _hidden_middle = modelling.NamedParentable("_Middle", parent=_hidden_top)
        _hidden_bottom = modelling.NamedParentable("_Bottom", parent=_hidden_middle)
        leaf = modelling.NamedParentable("Leaf", parent=_hidden_bottom)

    m = Model("TestModel")

    assert m.leaf.parent_chain == [
        m._hidden_bottom,
        m._hidden_middle,
        m._hidden_top,
        m,
    ]
    assert m._hidden_bottom.parent_chain == [
        m._hidden_middle,
        m._hidden_top,
        m,
    ]


def test_nested_models() -> None:
    """Ensure users are able to nest their own models."""

    class SubModel(modelling.NamedParentable):
        leaf = modelling.NamedParentable("Leaf")

    class Model(modelling.NamedParentable):
        sub = SubModel("TestSubModel")

    m = Model("TestModel")

    assert m.sub.parent == m
    assert m.sub.leaf.parent == m.sub


def test_clobbered_custom_models() -> None:
    """Ensure users are able to create their own models, even ignoring rules.

    While it is best to call `__init__`, etc. when extending frameworks, some
    don't. Additionally, sometimes they're simply called incorrectly. The
    framework is designed to be resilient to this abuse.

    This ensures the framework functions (delegates parents correctly), not
    necessarily that nice-to-haves (logging) work optimally.
    """

    class ClobberedModel(modelling.NamedParentable):
        # pylint: disable=super-init-not-called
        def __init__(self) -> None:
            self.custom_data = "foo"

        leaf = modelling.NamedParentable("Leaf")

    class Model(modelling.NamedParentable):
        clobbered = ClobberedModel()

    m = Model("TestModel")

    assert m.clobbered.leaf.parent_chain == [m.clobbered, m]


def test_manual_parent_override() -> None:
    """Ensure that users can override the parent in a `property`, if so desired.

    This case shouldn't be widely used, but should function as it is natural to
    assume it will work.
    """

    class Model(modelling.NamedParentable):
        _hidden = modelling.NamedParentable("Hidden")

        @property
        def leaf(self) -> modelling.NamedParentable:
            return modelling.NamedParentable("Leaf", parent=self._hidden)

    m = Model("TestModel")

    assert m.leaf.parent != m
    assert m.leaf.parent == m._hidden


def test_equal_definitions_are_separately_constructed() -> None:
    """Ensure otherwise equal children are both functional.

    This exists in case remapping via the metaclass ends up substituting (or
    leaving alone) an element in case of matching selectors and type (or the
    otherwise current definition of equality). Sometimes this is desired for
    logically different areas of the model that happen to have the same
    selector for now.

    Ensuring they are separate objects is best, since this ensures they were
    respected during construction. Equality is not the same, though.
    """

    class Model(modelling.NamedParentable):
        leaf1 = modelling.NamedParentable("Leaf")
        leaf2 = modelling.NamedParentable("Leaf")

    m = Model("TestModel")

    assert m.leaf1.parent == m
    assert m.leaf2.parent == m

    assert m.leaf1 is not m.leaf2


def test_equal_definitions_are_separately_constructed_middleman() -> None:
    """Same as previous test, but for deeper linkage."""

    class SubModel(modelling.NamedParentable):
        leaf = modelling.NamedParentable("Leaf")

    class Model(modelling.NamedParentable):
        sub1 = SubModel("Sub1")
        sub2 = SubModel("Sub2")

    m = Model("TestModel")

    assert m.sub1.leaf.parent == m.sub1
    assert m.sub2.leaf.parent == m.sub2


# FIXME: Test w/ user Model that injects another Metaclass (e.g. Generic[T])



"""Various tests to ensure dynamic (non-static) modelling is functional."""
# pylint: disable=missing-docstring,missing-class-docstring

from typing import Any
from unittest import mock

from e2e.common import modelling


def test_method_returned_modelling() -> None:
    """Ensure dynamic method-returned modelling works.

    Sometimes, you can't or don't want to make exact models for everything and
    they are parameterized (e.g. for list items). These are things you still
    want uniquely modelled, but the details are determined at test-time.

    Implicitly, this is expecting the framework to intercept the method's
    returned model to stitch the parent as the containing instance.
    """

    class SubModel(modelling.NamedParentable):
        leaf = modelling.NamedParentable("Leaf")

    class Model(modelling.NamedParentable):
        # pylint: disable=no-self-use  # intentional
        def sub_by_instance_method(self, label: str) -> SubModel:
            return SubModel("Sub by {}".format(label))

        @classmethod
        def sub_by_class_method(cls, label: str) -> SubModel:
            return SubModel("Sub by {}".format(label))

        @staticmethod
        def sub_by_static_method(label: str) -> SubModel:
            return SubModel("Sub by {}".format(label))

    m = Model("TestModel")

    assert m.sub_by_instance_method("foo-inst").parent == m
    assert m.sub_by_instance_method("bar-inst").parent == m

    assert (
        str(m.sub_by_instance_method("foo-inst").leaf)
        == "'Leaf' in 'Sub by foo-inst' in 'TestModel'"
    )
    assert (
        str(m.sub_by_instance_method("bar-inst").leaf)
        == "'Leaf' in 'Sub by bar-inst' in 'TestModel'"
    )

    # FIXME: Also test this with parent overrides

    # We can't support these; they don't make sense with the rest of how the
    # rest of the modelling framework works.
    assert m.sub_by_class_method("foo-cls").parent is None
    assert m.sub_by_class_method("bar-cls").parent is None
    assert m.sub_by_static_method("foo-static").parent is None
    assert m.sub_by_static_method("bar-static").parent is None


def test_other_methods_preserved() -> None:
    """Ensure other methods on models are not interrupted.

    Sometimes, it is desired to have helper methods on models to perform a
    series of actions or summarize some information. These should be left
    untouched so they operate as written.
    """

    class SubModel(modelling.NamedParentable):
        leaf = modelling.NamedParentable("Leaf")

    class Model(modelling.NamedParentable):
        sub = SubModel("TestSubModel")

        def interact_with_leaf(self) -> None:
            self.sub.leaf.do_action()  # type: ignore

        @property
        def ret_bool_prop(self) -> bool:
            return True

        # pylint: disable=no-self-use  # intentional
        def ret_bool_inst(self) -> bool:
            return True

        @classmethod
        def ret_bool_cls(cls) -> bool:
            return True

        @staticmethod
        def ret_bool_static() -> bool:
            return True

    # Access via class
    assert Model.ret_bool_cls() is True
    assert Model.ret_bool_static() is True

    m = Model("TestModel")

    # Access via instance
    assert m.ret_bool_prop is True
    assert m.ret_bool_inst() is True
    assert m.ret_bool_cls() is True
    assert m.ret_bool_static() is True

    m.sub.leaf.do_action = mock.Mock()  # type: ignore
    assert m.interact_with_leaf() is None  # type: ignore  # mypy over-assertion


def test_user_members_preserved() -> None:
    """Ensure other class/instance data on models are preserved.

    Sometimes, it is desired to have additional class members within a model
    for non-model data. These should not be touched by this framework, and
    continue to work as expected.
    """

    class SubModel(modelling.NamedParentable):
        leaf = modelling.NamedParentable("Leaf")

    class Model(modelling.NamedParentable):
        sub = SubModel("TestSubModel")
        class_data = "some class data"

        def __init__(self, *args: Any, **kwargs: Any) -> None:
            super().__init__(*args, **kwargs)
            self.instance_data = "some instance data"

    # Access via class
    assert Model.class_data == "some class data"

    m = Model("TestModel")

    # Access via instance
    assert m.sub.parent == m
    assert m.class_data == "some class data"
    assert m.instance_data == "some instance data"
